/* 
 * File:   MP3_15_26K22.c
 *
 * Created on 2021/02/12, 16:22
 */

#include <18F26K22.h>

#device ansi
#device const=read_only
#device PASS_STRINGS=IN_RAM
#opt compress

#fuses INTRC_IO,PLLEN,NOFCMEN,NOIESO,NOBROWNOUT,NOWDT,NOPBADEN,NOMCLR,STVREN,NOLVP,NOXINST,NOCPB,NOCPD,NODEBUG

//#use delay(clock=64000000)
#use delay(clock=32000000)
//#use delay(clock=16000000)
//#use delay(clock=8000000)

//#define PIC_CLOCK       OSC_64MHZ
#define PIC_CLOCK       OSC_32MHZ
//#define PIC_CLOCK       OSC_16MHZ
//#define PIC_CLOCK       OSC_8MHZ


#bit PEIE    = 0x0FF2.6       // INTCON.6

// ESP-WROOM-02との通信用
//#use rs232(STREAM=WIFI,BAUD=115200,UART2,PARITY=N,BITS=8,STOP=1,NOINIT,ERRORS)
#byte TXSTA2    = 0x0F72    // CSRC TX9 TXEN SYNC SENDB BRGH TRMT TX9D
#byte RCSTA2    = 0x0F71    // SPEN RX9 SREN CREN ADDEN FERR OERR RX9D
#byte TXREG2    = 0x0F73
#byte RCREG2    = 0x0F74
#byte BAUDCON2  = 0x0F70    // ABDOVF RCIDL DTRXP CKTXP BRG16 — WUE ABDEN
#byte SPBRGH2   = 0x0F76
#byte SPBRG2    = 0x0F75
#byte TRISB     = 0x0F93    // TRISB7 TRISB6 TRISB5 TRISB4 TRISB3 TRISB2 TRISB1 TRISB0
#bit TRISB7     = 0x0F93.7  // RX2
#bit TRISB6     = 0x0F93.6  // TX2
#bit TX2IF      = 0x0FA4.4  // PIR3.4
#bit RC2IF      = 0x0FA4.5  // PIR3.5
#bit RC2IE      = 0x0FA3.5  // PIE3.5
#bit CREN2      = 0x0F71.4
#bit FERR2      = 0x0F71.2
#bit OERR2      = 0x0F71.1
#bit TXEN2      = 0x0F72.5

// デバッグ用
#use rs232(STREAM=DEBUG,BAUD=9600,XMIT=PIN_A0,RCV=PIN_A1,FORCE_SW,PARITY=N,BITS=8,STOP=1,ERRORS)


#include <stdio.h>
#include <stdlib.h>

#include "integer.h"
#include "diskio.h"
#include "ffconf.h"
#include "ff.h"

#define DECODE_BUF_SIZE             512

// POWER
#define PIN_POWER_ON_MAIN       PIN_C1      //  (output)
#define PIN_DETECT_POWER        PIN_C0      // input
#define PIN_LED                 PIN_A5
#define PIN_WAKEUP_PIC          PIN_B0
#define PIN_DREQ_WIFI           PIN_A2
#define PIN_DETECT_SD           PIN_A3
#define PIN_OUTPUT_RELAY        PIN_A4
// VS1063
#define MP3_SO                  PIN_B2
#define MP3_DREQ                PIN_B4
#define MP3_SI                  PIN_B3
#define MP3_CS                  PIN_C6
#define MP3_DCS                 PIN_C7      // BSYNC
#define MP3_SCLK                PIN_B1
#define MP3_RESET               PIN_B5
//

// 音量
#define MIN_VOLUME              0x00        // 最大音量
#define MAX_VOLUME              0xD0        // 最小音量
#define DEFAULT_VOLUME          0x30
#define STEP_VOLUME             0x04

// 操作
#define PLAY_NEXT_MP3           0x91        // 同一ディレクトリ内の次のMP3ファイルへ
#define PLAY_OTHER_DIR          0x92        // ランダム再生において次のディスプレイへ
#define STOP_PLAY               0x93        // 再生停止
#define SHUTDOWN                0x94        // 電源オフ

// 時間定数
#define CONFIG_TIMEOUT          5       // 秒
#define NO_CTRLACCESS_TIMEOUT   (2*60)  // 待機中のコントローラからの操作待ちタイムアウト（秒）
#define WAITING_INTERVAL        2       // 秒
#define NOWPLAYING_INTERVAL     3       // 秒
#define NO_CTRLACTIVE_TIMEOUT   20      // 再生中のコントローラからのアクセス無しタイムアウト（秒）
#define CONTINOUS_PLAY_LIMIT    (3*60*60)   // 秒

// ファイルレイアウト（ディレクトリ構成）
#define DIRSTRUCTUREFILE                "/DIRLST15.DAT"
#define	RECLEN_DIRLIST					192UL		// レコード長
#define DL_DATATIMESTAMP_OFFSET         4           // 作成日時
#define DL_DATATIMESTAMP_LEN            14          // 作成日時
#define DL_DISPLAYNAME_LEN				32			// ディレクトリ表示名(SJIS)のバイト数
#define	MAXCHARS_DIRNAMEU				30			// ディレクトリ表示名(UNICODE)の最大文字数
#define	DL_DISPLAYNAMEU_LEN				62
#define DL_FULLDIR_OFFSET               94
#define DL_FULLDIR_LEN					61	
#define DL_FULLDIRBYTES_OFFSET          155
#define DL_FULLDIRBYTES_LEN				1
#define DL_DIRNAME_LEN					13
#define DL_RECNO_SUBDIR_START_LEN		2
#define DL_NUM_SUBDIR_LEN				1
#define DL_RECNO_PARENTDIR_LEN			2
#define DL_RECNO_GRANDPARENTDIR_LEN		2
#define DL_NUM_SNDFILE_OFFSET           176         // MP3ファイル数 オフセット
#define DL_NUM_SNDFILE_LEN				1
#define DL_RECNO_SNDFILE_OFFSET         177
#define DL_RECNO_SNDFILE_LEN			2
#define DL_FILLER_LEN					13

// ファイルレイアウト（MP3ファイルリスト）
#define SNDFILELISTFILE                 "/SNDLST15.DAT"
#define RECLEN_MP3LIST                  128UL
#define ML_FILENAME_OFFSET              106
#define ML_FILENAME_LEN                 13
#define ML_TIMESEC_OFFSET               119
#define ML_TIMESEC_LEN                  2
#define ML_FILESIZE_OFFSET              121
#define ML_FILESIZE_LEN                 4
#define ML_TAGSIZE_OFFSET               125
#define ML_TAGSIZE_LEN                  2

// ファイルレイアウト（ランダム再生順ファイル）
#define RANDOMORDERFILE                 "/RANDOM15.DAT"
#define RECLEN_RANDOMORDER              4
#define RO_UNIT_LEN                     1
#define RO_DIRRECNO_OFFSET              1
#define RO_DIRRECNO_LEN                 2
#define RO_TRACKNO_OFFSET               3
#define RO_TRACKNO_LEN                  1

// ファイルレイアウト（履歴ファイル）
#define HISTFILE                        "/HIST15.DAT"
#define RECLEN_HISTFILE                 11
#define HS_LASTDIR_OFFSET               0
#define HS_LASTDIR_LEN                  2
#define HS_LASTTRACK_OFFSET             2
#define HS_LASTTRACK_LEN                1
#define HS_LASTBYTE_OFFSET              3
#define HS_LASTBYTE_LEN                 4
#define HS_LASTSECOND_OFFSET            7
#define HS_LASTSECOND_LEN               2
#define HS_RANDOMRECNO_OFFSET           9
#define HS_RANDOMRECNO_LEN              2

// ボリュームファイル
#define VOLUMEFILE                      "VOLUME15.DAT"
#define RECLEN_VOLUME                   2UL

// ネットワーク設定
#define CONFIGFILE                      "/CONFIG15.DAT"
#define RECLEN_CONFIGFILE               78UL

// VS1063 patch
#define	PATHESFILE                      "/PATCH15.DAT"      // VS1063A PATCHES / Full Patch

#define WINCTRL_CMDLENGTH               16      // ソケットを通じて送受信する１単位のデータサイズ(BYTE)
#define WIFISRBUFSIZE                   (10*WINCTRL_CMDLENGTH)         // 64      // USARTの受信データのバッファサイズ(Byte)
#define FILERW_BUFSIZE                  256
#define BUFLEN_CONFIGFILE               (((RECLEN_CONFIGFILE - 1)/(WINCTRL_CMDLENGTH - 2)) + 1)*(WINCTRL_CMDLENGTH - 2)     // RECLEN_CONFIGFILEを(WINCTRL_CMDLENGTH - 2)の整数倍に切り上げ

#define WINCTRL_CMD_STARTUP         0x01         // コントローラ起動通知＆ホスト接続
#define WINCTRL_CMD_DATATIMESTAMP   0x02
#define WINCTRL_CMD_SHUTDOWN        0x03
#define WINCTRL_CMD_PLAYDIR         0x10
#define WINCTRL_CMD_PLAYONE         0x11
#define WINCTRL_CMD_PLAYRANDOM      0x12
#define WINCTRL_CMD_STOPPLAY        0x13
#define WINCTRL_CMD_NEXT            0x14
#define WINCTRL_CMD_RESUME          0x15
#define WINCTRL_CMD_VOLUP           0x16
#define WINCTRL_CMD_VOLDOWN         0x17
#define WINCTRL_CMD_FASTFORWARD     0x18
#define WINCTRL_CMD_REWIND          0x19
#define WINCTRL_CMD_CTRLACTIVE      0x40
#define WINCTRL_CMD_WAITING         0x41
#define WINCTRL_CMD_NOWPLAYING      0x42
#define WINCTRL_CMD_CONFIRM         0x43
#define WINCTRL_CMD_MESSAGE         0x80

#define MESSAGE_UNKNOWN_COMMAND         1
#define MESSAGE_RECEIVE_OVERRUN         2

// PIC・ESP-WROOM-02 間通信
#define ESPWROOM02_CMD_PICSTART         0x90    // 
#define ESPWROOM02_CMD_REQCONFIG        0x91    // ネットワーク設定情報要求 / 返信
#define ESPWROOM02_CMD_ACKCONFIG        0x92    // ネットワーク設定完了通知
#define ESPWROOM02_CMD_SERIALTIMEOUT    0x93    // タイムアウト(PIC→ESP02 で発生)
#define ESPWROOM02_CMD_RESETESP02       0x95    // ESP-WROOM-02 リセット要求
#define ESPWROOM02_CMD_STOPTOPIC        0x96    // Serial送信ブロック＆受信データクリア要求 / 返信
#define ESPWROOM02_CMD_RESTARTTOPIC     0x97    // Serial送信再開要求
#define ESPWROOM02_CMD_CHECKSUMERR      0x98    // CheckSumエラー通知

#define NOWPLAYING_FLAG_START       1       // 再生開始
#define NOWPLAYING_FLAG_HALFWAY     2       // 再生途中
#define NOWPLAYING_FLAG_END         3       // 再生終了

// for compatibility with XC8(stdint.h)
typedef unsigned int8   uint8_t;
typedef unsigned int16  uint16_t;
typedef unsigned int32  uint32_t;


// ESP-WROOM-02のネットワーク設定状況
enum NETCONFIG_STATUS {NETCONFIG_NOCONFIG, NETCONFIG_WAITING, NETCONFIG_OK, NETCONFIG_TIMEOUT};
// ESP-WROOM-02との通信同期
enum SYNC_WITH_ESP02_STATUS {SYNC_OK, SYNC_REQCONFIG, SYNC_RESTARTTOPIC};

static uint16_t     WaitMsCycle;
static uint16_t     Timer0InitValue;
static FATFS        Fatfs;
volatile int        TimerCounter;
volatile uint16_t   TimerSecond;                        // 秒カウンタ
static int          TimerStatus;
static uint16_t     ContinuousPlayingTime;
static BOOLEAN      ConnectedToCtrl = FALSE;            // 
static uint16_t     Volume;                             // 音量
static uint8_t      RecHist[RECLEN_HISTFILE];           // 履歴ファイルレコード
static uint8_t      RecConfig[BUFLEN_CONFIGFILE];       // ネットワーク設定ファイル
static uint8_t      DataTimestamp[DL_DATATIMESTAMP_LEN];
static uint8_t      Mp3Buff1[DECODE_BUF_SIZE];
volatile enum SYNC_WITH_ESP02_STATUS SyncWithEspFlag = SYNC_REQCONFIG;      // ESP-WROOM-02 -> PIC のUART通信　同期回復
static int          SyncWithEspCounter = 0;
static uint8_t      CmdSeq = 0;

// WIFIワーク変数
volatile uint8_t    WifiRcvBuf[WIFISRBUFSIZE];              // WIFI受信
volatile int16      RcvIdxW;        // 受信値書き込みインデックス
volatile int16      RcvIdxR;        // 受信値読み出しインデックス
volatile int16      NewBytes;       // WifiRcvBufの中の未処理データバイト数
uint8_t WifiSndBuf[WINCTRL_CMDLENGTH + 1];      // WIFI送信
//volatile int WifiSrIdx = 0;

const char UartSyncStr[13]={"MP3_15 START"};
const int UartSyncStrLen = 12;

void InitApp();
void SetPicClock(int16 level);
void SetPicTimer0(int in_play);
void wait_ms(uint16_t ms);
void SyncUartWithEsp02();
void PicGotoSleep();
void UartSendChar(char c);
void UartSendStr(char *str);
void UartSendData(uint8_t *data, uint8_t bytes);
void DebugDispSendData(uint8_t *data, uint8_t bytes);
uint8_t CalcCheckSum(uint8_t *data, uint8_t bytes);
void UartReadClear();
void UartReadAllClear();
void uart_interrupt();
int PlayDirStart(FIL *file1, FIL *file2, uint8_t *buf1, uint8_t *buf2, uint16_t dirrecno, uint8_t start_trackno, uint32_t start_byte, uint16_t start_second, BOOLEAN play_one_track, BOOLEAN random);
int PlayRandomStart(FIL *file1, FIL *file2, uint8_t *buf1, uint8_t *buf2);
int PlayOneMp3(FIL *file, const char *mp3file, uint16_t dirrecno, uint8_t trackno, uint32_t start_byte, uint16_t start_second, BOOLEAN random
        , uint32_t filesize, uint16_t tagsize, uint16_t total_second);
BOOLEAN ReadWifiConfigFile(FIL *file);
void SendWifiConfig(uint8_t *buf, int no);
BOOLEAN ReadDataTimestamp(FIL *file, uint8_t *buf);
void SendDataTimestamp();
void SendPicStart();
void SendStopToPic();
void SendRestartToPic();
void SendResetEsp02();
void SendWaiting();
void SendNowPlaying(uint16_t dirrecno, uint8_t trackno, uint32_t byteposition, uint16_t elapsed_time, uint8_t option_flag);
void SendAcceptShutdown();
void SendMessageUnknownCommand(uint8_t cmd);
void SendMessageReceiveOverrun();
FRESULT ReadRandomFile(FIL *Fil, uint8_t *buf, uint16_t *rndrecno);
BOOLEAN ReadHistory(FIL *Fil);
void SaveHistory(FIL *Fil);
void ReadVolumeFile(FIL *file, const char *volfile);
void SaveVolumeFile(FIL *file, const char *volfile);
void BlinkLed(int n, uint16_t on_ms, uint16_t off_ms);
// vs1063.c
void VS1063_HardwareReset();
void VS1063_SoftwareReset();
uint16_t VS1063_ReadSci(uint8_t addr);
void VS1063_WriteSci(uint8_t addr, uint16_t val);
void VS1063_SetVolume(uint16_t volume);
void VS1063_SendZeros(uint16_t cnt);
void LoadPluginFromFile(char *PluginFile, FIL *file);

#include "spi_sd.c"
#include "ff.c"
#include "vs1063.c"

/*
 * 
 */
int main(int argc, char** argv) {
    FRESULT     res;
    FIL         Fil1;
    FIL         Fil2;
    uint8_t     RecBuf1[FILERW_BUFSIZE];
    uint8_t     RecBuf2[FILERW_BUFSIZE];
    uint8_t     WorkBuf[WINCTRL_CMDLENGTH + 1];
    uint8_t     DebugBuf[WINCTRL_CMDLENGTH + 1];
    enum NETCONFIG_STATUS   ConfigStatus = NETCONFIG_NOCONFIG;
    int         ret;
    uint16_t    TimeoutStart;
    uint16_t    waiting_time;            // 操作待ち（再生停止中）
    uint16_t    ctrlactive_time;         // 
    BOOLEAN     SendWaitingAndReturn;
    BOOLEAN     force_waiting;          // タイマに関係なくWAITINGを送信
    int         wdtreset = 0;
    int         continue_flag = 1;
    uint8_t     csval;
    int         cnt;

    InitApp();
    fprintf(DEBUG, "MP3_15 start \r\n");
    
    if (input(PIN_DETECT_SD) == 1){     // SDが無い（プログラム書き換えの間だけ起動する）
        BlinkLed(1, 800, 800);
        fprintf(DEBUG, "No SD card \r\n");
        TimeoutStart = TimerSecond;
        if (input(PIN_DETECT_POWER) == 0){          // USB電源から起動のとき
            while (1){}
        } else {    // メイン電源から起動のとき
            while (1){
                // (ESP-WROOM-02用)PG書き換えのため、時間を置いてから停止する
                if ((TimerSecond - TimeoutStart) > 3*60) break;
            }
        }
        goto shutdown;
    }
    
    res = f_mount(&Fatfs, _T("0:"), 0);
    fprintf(DEBUG, "f_mount=%d\r\n", res);
    //if (res == FR_OK){    // マウントオプション 0 なのでチェック省略
    
    // ESP-WROOM-02のネットワーク設定をファイルから取得
    cnt = 2;
    do {
        if (ReadWifiConfigFile(&Fil1) == TRUE){
            break;
        }
        wait_ms(500);
    } while (--cnt > 0);
    if (cnt == 0){
        //BlinkLed(2, 800, 800);
        output_high(PIN_LED);
        wait_ms(3000);
        output_low(PIN_LED);
        fprintf(DEBUG, "ReadWifiConfigFile error\r\n");
        if (input(PIN_DETECT_POWER) == 0){          // USB電源から起動のとき
            while (1){}
        }
        goto shutdown;
    }

    // ESP-WROOM-02とのUART通信の同期
    SyncUartWithEsp02();
    // 同期できたことをESP-WROOM-02に伝える
    SendPicStart();
        
standby:

    ReadHistory(&Fil1);
    ReadDataTimestamp(&Fil1, RecBuf1);
    
    // コントローラからの操作待ち
    TimerSecond = 0;
    TimeoutStart = TimerSecond;
    waiting_time = 0;
    ctrlactive_time = 0;
    SendWaitingAndReturn = FALSE;
    force_waiting = FALSE;
    continue_flag = 1;
    
    while (continue_flag){
        if (NewBytes >= WINCTRL_CMDLENGTH){
            memcpy(WorkBuf, WifiRcvBuf + RcvIdxR, WINCTRL_CMDLENGTH);
            RcvIdxR += WINCTRL_CMDLENGTH;
            if (RcvIdxR == WIFISRBUFSIZE) RcvIdxR = 0;
            NewBytes -= WINCTRL_CMDLENGTH;
            
            // デバッグ表示ここから
            if (wdtreset == 0 && WorkBuf[0] == 0x0d && WorkBuf[1] == 0x0a) wdtreset = 1;
            if (wdtreset == 0){
                fprintf(DEBUG, "Received ");
                for (int i=0; i<WINCTRL_CMDLENGTH; i++){
                    fprintf(DEBUG, "(%x)", WorkBuf[i]);
                }
                fprintf(DEBUG, " \r\n");
            } else {
                for (int i=0; i<WINCTRL_CMDLENGTH; i++){
                    DebugBuf[i] = WorkBuf[i];
                }
                DebugBuf[WINCTRL_CMDLENGTH] = '\0';
                fprintf(DEBUG, "%s", DebugBuf);
                fprintf(DEBUG, " \r\n");
            }
            // デバッグ表示ここまで
            
            csval = CalcCheckSum(WorkBuf, WINCTRL_CMDLENGTH - 1);
            if (csval != WorkBuf[WINCTRL_CMDLENGTH - 1]){
                fprintf(DEBUG, "CheckSum error ESP02->PIC \r\n");
            }
            
            if (WorkBuf[0] == WINCTRL_CMD_STARTUP){                 // コントローラ起動通知
                fprintf(DEBUG, "Connect\r\n");
                SendDataTimestamp();
            } else if (WorkBuf[0] == WINCTRL_CMD_DATATIMESTAMP){    // データファイル　タイムスタンプ要求
                fprintf(DEBUG, "Get Datetimestamp Cmd\r\n");
                SendDataTimestamp();
            } else if (WorkBuf[0] == WINCTRL_CMD_SHUTDOWN){         // コントローラ切断＆ホストオフ
                ConnectedToCtrl = FALSE;
                SendAcceptShutdown();
                continue_flag = 0;
            } else if (WorkBuf[0] == WINCTRL_CMD_PLAYDIR){          // 指定ディレクトリの特定曲番号から再生開始
                uint16_t    recno = WorkBuf[1] | ((uint16_t)WorkBuf[2] << 8);
                uint8_t     start_trackno = WorkBuf[3];
                uint32_t    start_byte = 0;     // 不明なのでPlayDirStart内で計算　 = WorkBuf[4] | ((uint32_t)WorkBuf[5] << 8)  | ((uint32_t)WorkBuf[6] << 16)  | ((uint32_t)WorkBuf[7] << 24);
                uint16_t    start_second = WorkBuf[4] | ((uint32_t)WorkBuf[5] << 8);
                ContinuousPlayingTime = TimerSecond;
                //SetPicTimer0(1);
                ret = PlayDirStart(&Fil1, &Fil2, RecBuf1, RecBuf2, recno, start_trackno, start_byte, start_second, FALSE, FALSE);
                if (ret == SHUTDOWN){
                    continue_flag = 0;
                }
                //SetPicTimer0(0);
                TimeoutStart = TimerSecond;
                SendWaitingAndReturn = FALSE;
                force_waiting = TRUE;
            } else if (WorkBuf[0] == WINCTRL_CMD_RESUME){           // 前回再生の再開
                uint16_t    recno = RecHist[HS_LASTDIR_OFFSET] | ((uint16_t)RecHist[HS_LASTDIR_OFFSET+1] << 8);
                uint8_t     start_trackno= RecHist[HS_LASTTRACK_OFFSET];
                uint32_t    start_byte = RecHist[HS_LASTBYTE_OFFSET] | ((uint32_t)RecHist[HS_LASTBYTE_OFFSET+1] << 8)  | ((uint32_t)RecHist[HS_LASTBYTE_OFFSET+2] << 16)  | ((uint32_t)RecHist[HS_LASTBYTE_OFFSET+3] << 24);
                uint16_t    start_second = RecHist[HS_LASTSECOND_OFFSET] | ((uint16_t)RecHist[HS_LASTSECOND_OFFSET+1] << 8);
                ContinuousPlayingTime = TimerSecond;
                //SetPicTimer0(1);
                ret = PlayDirStart(&Fil1, &Fil2, RecBuf1, RecBuf2, recno, start_trackno, start_byte, start_second, FALSE, FALSE);
                if (ret == SHUTDOWN){
                    continue_flag = 0;
                }
                //SetPicTimer0(0);
                TimeoutStart = TimerSecond;
                SendWaitingAndReturn = FALSE;
                force_waiting = TRUE;
            } else if (WorkBuf[0] == WINCTRL_CMD_PLAYRANDOM){       // ランダム再生開始
                ContinuousPlayingTime = TimerSecond;
                //SetPicTimer0(1);
                ret = PlayRandomStart(&Fil1, &Fil2, RecBuf1, RecBuf2);
                if (ret == SHUTDOWN){
                    continue_flag = 0;
                }
                //SetPicTimer0(0);
                TimeoutStart = TimerSecond;
                SendWaitingAndReturn = FALSE;
                force_waiting = TRUE;
            } else if (WorkBuf[0] == WINCTRL_CMD_CTRLACTIVE){       // コントローラ実行中であることを通知
                //fprintf(DEBUG, "CtrlActive \r\n");
                SendWaitingAndReturn = FALSE;       // WAITING送信に対する返信
                ctrlactive_time = TimerSecond;
                ConnectedToCtrl = TRUE;
            } else if (WorkBuf[0] == WINCTRL_CMD_CONFIRM){          // HOST動作確認
                UartSendData(WorkBuf, WINCTRL_CMDLENGTH);
                DebugDispSendData(WorkBuf, WINCTRL_CMDLENGTH);
            } else if (WorkBuf[0] == ESPWROOM02_CMD_REQCONFIG){     // ESP02からネットワーク設定情報要求
                SendWifiConfig(RecBuf1, WorkBuf[1]);                // 同　返信
            } else if (WorkBuf[0] == ESPWROOM02_CMD_ACKCONFIG){     // ネットワーク設定完了通知
                fprintf(DEBUG, "Network complete (%ld)\r\n", TimerSecond);
                output_high(PIN_LED);
                wait_ms(500);
                output_low(PIN_LED);
                ConfigStatus = NETCONFIG_OK;
                if (input(PIN_DETECT_POWER) == 0){          // メイン電源起動していない。USB電源から起動のとき
                    PicGotoSleep();
                }
                output_high(PIN_LED);
                VS1063_HardwareReset();
                //LoadPluginFromFile(PATHESFILE, &Fil1);
            } else if (WorkBuf[0] == ESPWROOM02_CMD_CHECKSUMERR){
                fprintf(DEBUG, "CheckSum report from ESP02\r\n");
                TXEN2 = 0;
                SendWaitingAndReturn = FALSE;
                ctrlactive_time = TimerSecond;
                // 送信
                TXEN2 = 1;
            } else if (WorkBuf[0] == ESPWROOM02_CMD_SERIALTIMEOUT){
                fprintf(DEBUG, "TIMEOUT PIC->ESP02\r\n");
            } else {
                UartReadAllClear();
                if (wdtreset == 0){
                    SendMessageUnknownCommand(WorkBuf[0]);
                }
            }
            
        }
        
        if (ConfigStatus == NETCONFIG_OK){
            if (ConnectedToCtrl == TRUE){
                if (SendWaitingAndReturn == FALSE && ((TimerSecond - ctrlactive_time) > WAITING_INTERVAL || force_waiting == TRUE)){
                    SendWaiting();
                    SendWaitingAndReturn = TRUE;
                    waiting_time = TimerSecond;
                    force_waiting = FALSE;
                } else if (SendWaitingAndReturn == TRUE && (TimerSecond - waiting_time) > NO_CTRLACCESS_TIMEOUT){
                    // WAITINGの返信が無い
                    fprintf(DEBUG, "No response for Waiting \r\n");
                    ConnectedToCtrl = FALSE;        // コントローラの接続が切れた
                    TimeoutStart = TimerSecond;
                }
            } else {
                if ((TimerSecond - TimeoutStart) > NO_CTRLACCESS_TIMEOUT){         // コントローラからのアクセスが無いとき終了
                    BlinkLed(3, 800, 800);
                    fprintf(DEBUG, "No Ctrl Access Timeout (%ld)(%ld) \r\n", TimerSecond, TimeoutStart);
                    break;
                }
            }
        } else {        // ESP-WROOM-02のネットワーク設定されていない
            if ((TimerSecond - TimeoutStart) > CONFIG_TIMEOUT){
                BlinkLed(2, 800, 800);
                fprintf(DEBUG, "Netconfig Failed timeout (%ld)(%ld) \r\n", TimerSecond, TimeoutStart);
                break;
            }
        }
    }
    
    SaveHistory(&Fil1);
shutdown:
    ConnectedToCtrl = FALSE;
    output_low(PIN_LED);
    wait_ms(500);          // コントローラに通知する時間
    output_low(PIN_POWER_ON_MAIN);      // メイン電源遮断
    // USB電源があれば以下続行（のはずだが、実際にはメイン電源→USB電源の移行はスムーズに行われなくて再起動してしまう）
    wait_ms(1000);
    PicGotoSleep();
    output_high(PIN_LED);
    VS1063_HardwareReset();
    //LoadPluginFromFile(PATHESFILE, &Fil1);
    goto standby;
    
//    } else {
//        //fprintf(DEBUG, "f_mount=%d\r\n", res);
//    }
    
    while (1){
        
    }
    return (0);
}


void InitApp(){
    //FRESULT     res;
    
    output_low(PIN_POWER_ON_MAIN);
    output_float(PIN_DETECT_POWER);
    output_float(PIN_WAKEUP_PIC);
    output_float(PIN_DETECT_SD);
    output_low(PIN_OUTPUT_RELAY);       // リレーオフ
    output_high(MP3_DCS);
    output_high(MP3_CS);
    output_high(MP3_RESET);
    output_float(MP3_DREQ);   // DREQ
//    port_b_pullups(0x10);     // B4
    
    output_low(PIN_LED);
    
    /* Initialize peripherals */
    setup_adc(ADC_OFF);         // ADC disabled
    setup_ccp1(CCP_OFF);        // CCP1 disabled
    setup_ccp2(CCP_OFF);        // CCP2 disabled

    if (input(PIN_DETECT_POWER) == 1){
        output_high(PIN_POWER_ON_MAIN);
    }
    
    SetPicClock(PIC_CLOCK);
    
    // USART2の設定(32MHz) ボーレート115200
    // X = ((Fosc/Desired Baud Rate)/4) – 1
    //   = ((32000000/115200) / 4) - 1
    //   = 68.444
    // USART2の設定(32MHz) ボーレート9600
    //   = ((32000000/9600) / 4) - 1
    //   = 832.333 = 0x0340
    // USART2の設定(8MHz) ボーレート9600
    //   = ((8000000/9600) / 4) - 1
    //   = 207.333 = 0x00CF
    // 
    RcvIdxW = 0;
    RcvIdxR = 0;
    NewBytes = 0;
    TRISB7 = 1;
    TRISB6 = 1;
    TXSTA2 = 0b00100100;    // 送信情報設定：非同期モード 8ビットデータ ノンパリティ //SYNC=0;BRGH=1;
    RCSTA2 = 0b10010000;    // 受信情報設定：Enable、８Bit、Continuous Receive Enable
    BAUDCON2 = 0b01001000;  // BRG16=1;    Receive operation is Idle、16-bit Baud Rate Generator – SPBRGHx and SPBRGx
    switch (PIC_CLOCK){
    case OSC_64MHZ:
        SPBRGH2 = 0x06;    // 16-bit Baud Rate Generator
        SPBRG2 = 0x82;    // 16-bit Baud Rate Generator
        break;
    case OSC_32MHZ:
        SPBRGH2 = 0x03;    // 16-bit Baud Rate Generator
        SPBRG2 = 0x40;    // 16-bit Baud Rate Generator
        //↓115200
        //SPBRGH2 = 0;    // 16-bit Baud Rate Generator
        //SPBRG2 = 68;    // 16-bit Baud Rate Generator
        break;
    case OSC_16MHZ:
        SPBRGH2 = 0x01;    // 16-bit Baud Rate Generator
        SPBRG2 = 0xA0;    // 16-bit Baud Rate Generator
        break;
    case OSC_8MHZ:
        SPBRGH2 = 0x00;    // 16-bit Baud Rate Generator
        SPBRG2 = 0xCF;    // 16-bit Baud Rate Generator
        break;
    }
    RC2IF = 0;      // EUSART2 Receive Interrupt Flag bit
    RC2IE = 1;      // Enables the EUSART2 receive interrupt

    PEIE = 1;
    
    fprintf(DEBUG, "InitApp()\r\n");
    
    Volume = DEFAULT_VOLUME;    
//    VS1063_HardwareReset();
    
    TimerCounter = 0;
    TimerSecond = 0;
    SetPicTimer0(0);
//    setup_timer_0(T0_INTERNAL | T0_DIV_16);
//    set_timer0(Timer0InitValue);
//    clear_interrupt(INT_TIMER0);
//    enable_interrupts(INT_TIMER0);

    enable_interrupts(GLOBAL);
    
//    wait_ms(1000);
//    UartReadAllClear();    // ESP-WROOM-02 起動メッセージを読み飛ばす

}

/*
 *  PICのクロック設定
 */
void SetPicClock(int16 level){
    switch (level){
    case OSC_64MHZ:
        setup_oscillator(OSC_64MHZ);
        
        // (1000us)/(4/64MHz)*250 = 64
        // -> 64*delay_cycle(250) = 1ms
        WaitMsCycle = 64;
        
        // Timer0 : (4/64MHz) * 32(prescaler) = 2us
        // 2us * (65536 - 15536) = 100ms
        Timer0InitValue = 15536;
        
        break;
            
    case OSC_32MHZ:    
        setup_oscillator(OSC_32MHZ);
        
        // (1000us)/(4/32MHz)*250 = 32
        // -> 32*delay_cycle(250) = 1ms
        WaitMsCycle = 32;
        
        // Timer0 : (4/32MHz) * 16(prescaler) = 2us
        // 2us * 65536 = 131ms
        //Timer0InitValue = 0;
        // 2us * (65536 - 15536) = 100ms
        Timer0InitValue = 15536;
        
        // Timer1 : (4/32MHz) * 8(prescaler) = 1 us
        // 1us * 65536 = 65ms
        //Timer1InitValue = 0;
        
        // Timer0 : (4/32MHz) * 8(prescaler) = 1 us
        //Timer0InitValue = (65536 - 62500);
        
        break;
    case OSC_16MHZ:
        setup_oscillator(OSC_16MHZ);
        
        // (1000us)/(4/16MHz)*250 = 16
        // -> 16*delay_cycle(250) = 1ms
        WaitMsCycle = 16;
        
        // Timer0 : (4/16MHz) * 16(prescaler) = 4us
        // 4us * 65536 = 262ms
        // 4us * (65536 - 40536) = 100ms
        Timer0InitValue = 40536;
        
        break;
    case OSC_8MHZ:
        setup_oscillator(OSC_8MHZ);
        
        // (1000us)/(4/8MHz)*250 = 8
        // -> 8*delay_cycle(250) = 1ms
        WaitMsCycle = 8;
        
        // Timer0 : (4/8MHz) * 16(prescaler) = 8us
        // 8us * 65536 = 254ms
        // 8us * (65536 - 53036) = 100ms
        Timer0InitValue = 53036;
        
        // Timer1 : (4/8MHz) * 8(prescaler) = 4 us
        // 4us * (65536/4) = 65ms
        //Timer1InitValue = 49152;    // = 65536 - (65536 / 4)
        
        // Timer0 : (4/8MHz) * 8(prescaler) = 4 us
        // 128 us * 15625 = 2s
        //Timer0InitValue = (65536 - 15625);
        
        break;
    case OSC_1MHZ:
        setup_oscillator(OSC_1MHZ);
        
        // (1000us)/(4/1MHz)*250 = 1
        // -> 1*delay_cycle(250) = 1ms
        WaitMsCycle = 1;
        
        // Timer0 : (4/1MHz) * 16(prescaler) = 64us
        // 64us * (65536/32) = 131ms
        Timer0InitValue = 63488;    // = 65536 - (65536 / 32)
        
        // Timer1 : (4/1MHz) * 8(prescaler) = 32 us
        // 32us * (65536/32) = 65ms
        //Timer1InitValue = 63488;    // = 65536 - (65536 / 32)
        
        // Timer0 : (4/1MHz) * 8(prescaler) = 32 us
        // 1024 us * 1953 = 2s
        //Timer0InitValue = (65536 - 1953);
        
        break;
    }
    
    //setup_uart(9600, DEBUG);
    
    wait_ms(1000);
}

/*
 *  TIMER0設定
 *      in_playでタイマー割り込み周期指定
 *      in_play==0:100ms
 *      in_play==1:1000ms
 */
void SetPicTimer0(int in_play){
    if (in_play == 0){  // out of play
        TimerStatus = 0;
        disable_interrupts(INT_TIMER0);
        
        // Timer0 : (4/32MHz) * 16(prescaler) = 2us
        // 2us * 65536 = 131ms
        //Timer0InitValue = 0;
        // 2us * (65536 - 15536) = 100ms
        Timer0InitValue = 15536;
        setup_timer_0(T0_INTERNAL | T0_DIV_16);

        set_timer0(Timer0InitValue);
        clear_interrupt(INT_TIMER0);
        enable_interrupts(INT_TIMER0);
        
    } else {        // in play
        TimerStatus = 1;        
        disable_interrupts(INT_TIMER0);
        
        // Timer0 : (4/32MHz) * 128(prescaler) = 16us
        // 16us * 65536 = 1049ms
        //Timer0InitValue = 0;
        // 16us * (65536 - 3036) = 1000ms
        Timer0InitValue = 3036;
        setup_timer_0(T0_INTERNAL | T0_DIV_128);

        set_timer0(Timer0InitValue);
        clear_interrupt(INT_TIMER0);
        enable_interrupts(INT_TIMER0);
        
    }
}


/*
 *  ESP-WROOM-02とのUART通信の同期
 *  ESP-WROOM-02から送られてくる所定文字列を検出する
 */
void SyncUartWithEsp02(){
    int16   idxw;
    int16   i, j;
    BOOLEAN flag;
    
    while (1){
        flag = TRUE;
        idxw = RcvIdxW;
        if (idxw == 0) idxw = WIFISRBUFSIZE;
        for (i=1,j=idxw-1; i<=UartSyncStrLen; i++){
            fprintf(DEBUG, "%ld %c %c\r\n", i, WifiRcvBuf[j], UartSyncStr[UartSyncStrLen - i]);
            if (WifiRcvBuf[j] != UartSyncStr[UartSyncStrLen - i]){
                flag = FALSE;
                //break;
            }
            j--;
            if (j < 0) j = WIFISRBUFSIZE - 1;
        }
        
        if (flag == TRUE){
            RcvIdxW = 0;
            NewBytes = 0;
            SyncWithEspFlag = SYNC_OK;
            fprintf(DEBUG, "SYNC OK\r\n");
            break;
        } else {
            fprintf(DEBUG, "SYNC NG\r\n");
        }
        wait_ms(500);
    }
}

/*
 *  PICをSLEEP状態に。WAKEUPの条件設定も行う。
 */
void PicGotoSleep(){
    fprintf(DEBUG, "PIN_DETECT_POWER low->sleep\r\n");
    output_low(PIN_LED);
    // SLEEPに入る
    while (input(PIN_WAKEUP_PIC) != 1);
    fprintf(DEBUG, "PIN_DETECT_POWER low->sleep\r\n");
    ext_int_edge(0, H_TO_L);      // init interrupt triggering for button press
    enable_interrupts(INT_EXT_H2L);// turn on interrupts
    enable_interrupts(GLOBAL);
    sleep();
    output_high(PIN_POWER_ON_MAIN);
    fprintf(DEBUG, "WAKEUP\r\n");
}


void wait_ms(uint16_t ms){
    uint8_t     i;
    uint16_t    m;
    
    for (m=0; m<ms; m++){
        
        // 1ms
        for (i=0; i<WaitMsCycle; i++){
            delay_cycles(250);
        }
    }
}

// USART 1文字送信
void UartSendChar(char c){
    while(TX2IF == 0) ;//送信可能まで待つ
    TXREG2 = c;
}

// WIFI文字列送信
void UartSendStr(char *str){
    int     i;
    
    for (i=0; str[i] != '\0'; i++){
        UartSendChar(str[i]);
    }
}

// WIFIデータ送信
void UartSendData(uint8_t *data, uint8_t bytes){
    uint8_t     i;
    
    for (i=0; i<bytes; i++){
        UartSendChar(data[i]);
    }
}

void DebugDispSendData(uint8_t *data, uint8_t bytes){
    fprintf(DEBUG, "Sent ");
    for (int i=0; i<bytes; i++){
        fprintf(DEBUG, "(%x)", data[i]);
    }
    fprintf(DEBUG, " \r\n");
}

// Checksum計算
uint8_t CalcCheckSum(uint8_t *data, uint8_t bytes){
    uint16_t    cs = 0;
    int         i;
    
    for (i=0; i<bytes; i++){
        cs += data[i];
    }
    
    return (cs & 0xFF);
}


// 
void UartReadClear(){
    int     c;
    
    while (RC2IF == 1){
        c = RCREG2;
    }
}

// ESP-WROOM-02の起動時メッセージを飛ばす
void UartReadAllClear(){
    int     c;
    
    while (1){
        while (RC2IF == 1){
            c = RCREG2;
        }
        wait_ms(10);       // しばらく待って次の入力が無ければすべて読んだ
        if (RC2IF == 0) break;
    }
}

#int_rda2
void uart_interrupt(){
    int     c;
    
    if (RC2IF == 1){
        c = RCREG2;
        if (SyncWithEspFlag == SYNC_OK){
            if (NewBytes < WIFISRBUFSIZE){
                WifiRcvBuf[RcvIdxW++] = c;
                NewBytes++;
                if (RcvIdxW == WIFISRBUFSIZE) RcvIdxW = 0;
            }
        } else if (SyncWithEspFlag == SYNC_REQCONFIG){
                WifiRcvBuf[RcvIdxW++] = c;
                NewBytes++;
                if (RcvIdxW == WIFISRBUFSIZE) RcvIdxW = 0;
        }
    }
}

/*
 *  ディレクトリ構成ファイルのタイムスタンプ取得
 */
BOOLEAN ReadDataTimestamp(FIL *file, uint8_t *buf){
    FRESULT     res;
    UINT        br;
    //int         i;
    BOOLEAN     ret = FALSE;
    
    res = f_open(file, DIRSTRUCTUREFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_read(file, buf, RECLEN_DIRLIST, &br);       // ヘッダレコード
        if (res == FR_OK && br == RECLEN_DIRLIST){
            memcpy(DataTimestamp, buf + DL_DATATIMESTAMP_OFFSET, DL_DATATIMESTAMP_LEN);
            ret = TRUE;
        }
        f_close(file);
    }
    
    return ret;
}


/*
 *  ディレクトリ構成ファイルのタイムスタンプをコントローラに送る
 */
void SendDataTimestamp(){
    int         i;
    
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = WINCTRL_CMD_DATATIMESTAMP;
    for (i=0; i<DL_DATATIMESTAMP_LEN; i++){
        WifiSndBuf[1 + i] = DataTimestamp[i];
    }
    // SEQ番号を付けない
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
}


/*
 *  指定ディレクトリの特定曲番号から再生開始
 *      start_byte : 最初の曲の再生開始バイト位置
 *      start_second : 最初の曲の再生開始時刻(秒)
 *      play_one_track : 特定曲番号のみ再生のときTRUE
 *      random : ランダム再生のときTRUE
 *
 */
int PlayDirStart(FIL *file1, FIL *file2, uint8_t *buf1, uint8_t *buf2, uint16_t dirrecno, uint8_t start_trackno, uint32_t start_byte
        , uint16_t start_second, BOOLEAN play_one_track, BOOLEAN random){
    FRESULT     res;
    UINT        br;
    uint8_t     fulldirbytes;
    uint16_t    num_sndfile;
    uint16_t    recno_sndfile;
    uint16_t    trackno;
    uint16_t    tagsize;
    uint16_t    total_second;
    uint32_t    filesize;
    //uint32_t    bytespersec;            // 1秒あたりの平均バイト数
    //int         i;
    int         ret;
    
    ret = FR_OK;
    
    // ディレクトリ構成ファイルレコード取得
    res = f_open(file1, DIRSTRUCTUREFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        fprintf(DEBUG, "dirfile f_open=%d\r\n", res);
        return res;
    }
    res = f_lseek(file1, (uint32_t)dirrecno * RECLEN_DIRLIST);
    if (res != FR_OK){
        fprintf(DEBUG, "dirfile f_lseek=%d\r\n", res);
        return res;
    }
    res = f_read(file1, buf1, RECLEN_DIRLIST, &br);
    if (res != FR_OK){
        fprintf(DEBUG, "dirfile f_read=%d\r\n", res);
        return res;
    }
    f_close(file1);
    
    fulldirbytes = *(buf1 + DL_FULLDIRBYTES_OFFSET);
    num_sndfile = *(buf1 + DL_NUM_SNDFILE_OFFSET);
    recno_sndfile = *((uint16_t *)(buf1 + DL_RECNO_SNDFILE_OFFSET));
    if (num_sndfile == 0) return FR_OK;
    
    // 以降、buf1はMP3ファイルまたはボリュームファイルのフルパス作業用に使う
    
    // ボリュームファイル
    *(buf1 + DL_FULLDIR_OFFSET + fulldirbytes) = '/';
    strcpy(buf1 + DL_FULLDIR_OFFSET + fulldirbytes + 1, VOLUMEFILE);
    ReadVolumeFile(file1, buf1 + DL_FULLDIR_OFFSET);
    VS1063_SetVolume(Volume);
    
    // MP3ファイル
    res = f_open(file1, SNDFILELISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        fprintf(DEBUG, "sndfile f_open=%d\r\n", res);
        return res;
    }
    res = f_lseek(file1, (uint32_t)(recno_sndfile + start_trackno) * RECLEN_MP3LIST);
    if (res != FR_OK){
        fprintf(DEBUG, "sndfile f_lseek=%d\r\n", res);
        return res;
    }
    
    output_high(PIN_OUTPUT_RELAY);       // リレーオン
    for (trackno=start_trackno; trackno<num_sndfile; trackno++){
        res = f_read(file1, buf2, RECLEN_MP3LIST, &br);
        if (res != FR_OK){
            fprintf(DEBUG, "sndfile f_read=%d\r\n", res);
            break;
        }
        tagsize = (uint16_t)buf2[ML_TAGSIZE_OFFSET] | (uint16_t)(buf2[ML_TAGSIZE_OFFSET + 1] << 8);
        memcpy(&total_second, buf2 + ML_TIMESEC_OFFSET, ML_TIMESEC_LEN);
        memcpy(&filesize, buf2 + ML_FILESIZE_OFFSET, ML_FILESIZE_LEN);
        if (trackno > start_trackno){
            start_byte = tagsize;
            start_second = 0;
        } else {    // 最初の曲(start_trackno)の再生
            if (start_second > 0 && start_byte == 0){   // 開始秒に対応する開始バイト位置が不明
                if (start_second < total_second){
                    start_byte = tagsize + (uint32_t)((filesize - tagsize) * ((float)start_second / (float)total_second));
                } else {
                    start_byte = filesize;
                }
                fprintf(DEBUG, "sndfile start_sec=%ld total_sec=%ld start_byte=%ld\r\n", start_second, total_second, start_byte);
            } else if (start_second == 0 && start_byte == 0){
                start_byte = tagsize;
            }
        }
        // MP3ファイルのフルパス
        *(buf1 + DL_FULLDIR_OFFSET + fulldirbytes) = '/';
        strcpy(buf1 + DL_FULLDIR_OFFSET + fulldirbytes + 1, buf2 + ML_FILENAME_OFFSET);
        ret = PlayOneMp3(file2, buf1 + DL_FULLDIR_OFFSET, dirrecno, trackno, start_byte, start_second, random, filesize, tagsize, total_second);
        if (ret == PLAY_NEXT_MP3){
            continue;
        } else if (ret == PLAY_OTHER_DIR){      // ランダム再生で次のディレクトリへ
            break;
        } else if (ret == STOP_PLAY){
            break;
        } else if (ret == SHUTDOWN){
            break;
        }
        if (play_one_track == TRUE) break;
    }
    output_low(PIN_OUTPUT_RELAY);       // リレーオフ
    
    f_close(file1);
    
    // ボリュームファイル
    fprintf(DEBUG, "Volume file save start \r\n");
    *(buf1 + DL_FULLDIR_OFFSET + fulldirbytes) = '/';
    strcpy(buf1 + DL_FULLDIR_OFFSET + fulldirbytes + 1, VOLUMEFILE);
    SaveVolumeFile(file1, buf1 + DL_FULLDIR_OFFSET);
    fprintf(DEBUG, "Volume file saved \r\n");
    
    return ret;
}

// ランダム再生開始
int PlayRandomStart(FIL *file1, FIL *file2, uint8_t *buf1, uint8_t *buf2){    
    FRESULT     res;
    uint16_t    rndrecno;
    uint16_t    dirrecno;
    uint8_t     trackno;
    int         ret;
    
    ret = FR_OK;
    memcpy(&rndrecno, RecHist + HS_RANDOMRECNO_OFFSET, HS_RANDOMRECNO_LEN);
    
    while (1){
        res = ReadRandomFile(file1, buf1, &rndrecno);
        if (res != FR_OK) break;
        
        memcpy(&dirrecno, buf1 + RO_DIRRECNO_OFFSET, RO_DIRRECNO_LEN);
        memcpy(&trackno, buf1 + RO_TRACKNO_OFFSET, RO_TRACKNO_LEN);
        
        if (buf1[0] == 0){      // フォルダ単位
            ret = PlayDirStart(file1, file2, buf1, buf2, dirrecno, 0, 0, 0, FALSE, TRUE);
            if (ret == STOP_PLAY){
                break;
            } else if (ret == SHUTDOWN){
                break;
            }
        } else {    // ファイル単位
            ret = PlayDirStart(file1, file2, buf1, buf2, dirrecno, trackno, 0, 0, TRUE, TRUE);
            if (ret == STOP_PLAY){
                break;
            } else if (ret == SHUTDOWN){
                break;
            }
        }
        rndrecno++;
    }
    
    rndrecno++;     // 次回のランダム再生は今回最後のディレクトリの次から
    memcpy(RecHist + HS_RANDOMRECNO_OFFSET, &rndrecno, HS_RANDOMRECNO_LEN);
    
    return ret;
}


/*
 *  MP3ファイル一つ再生
 */
int PlayOneMp3(FIL *file, const char *mp3file, uint16_t dirrecno, uint8_t trackno, uint32_t start_byte, uint16_t start_second, BOOLEAN random
        , uint32_t filesize, uint16_t tagsize, uint16_t total_second){
    FRESULT     res;
    UINT        sendbytes, sentbytes;
    UINT        bytes;
    int         continue_flag, ret, i;
    uint32_t    read_byte;              // MP3ファイルの読み出しバイト位置
    uint16_t    elapsed_time;           // 曲の再生経過時間（タイマイベントでリアルタイムに更新されない。必要な時に再計算する）
    uint16_t    elapsed_time_start;     // 
    uint16_t    timer_second_base;      // 
    BOOLEAN     send_nowplaying;        // NOWPLAYINGを送信して返答待ち
    BOOLEAN     force_nowplaying;       // タイマに関係なくNOWPLAYINGを送信
    uint16_t    ctrlactive_time;        // CTRLACTIVE受信のタイマ時刻
    uint8_t     WorkBuf[WINCTRL_CMDLENGTH + 1];
    uint8_t     csval;          // check sum 値
    uint16_t    skiptime;
    uint32_t    bytespersec;    // 1秒あたりの平均バイト数
    DWORD       ofs;
    DWORD       ofs_bk;

    ret = FR_OK;
    fprintf(DEBUG, "PlayOneMp3 start \r\n");
    
    res = f_open(file, mp3file, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        fprintf(DEBUG, "mp3file f_open=%d\r\n", res);
        return res;
    }
    read_byte = start_byte;
    if (start_byte > 0){
        res = f_lseek(file, start_byte);
    }
    elapsed_time = start_second;
    elapsed_time_start = start_second;
    timer_second_base = TimerSecond;
    // 以後、elapsed_time = TimerSecond - timer_second_base + elapsed_time_start　で計算される
    if (total_second > 0){
        bytespersec = (filesize - tagsize) / total_second;
    } else {
        bytespersec = 0;
    }
    
    memcpy(RecHist + HS_LASTDIR_OFFSET, &dirrecno, HS_LASTDIR_LEN);
    memcpy(RecHist + HS_LASTTRACK_OFFSET, &trackno, HS_LASTTRACK_LEN);
    memcpy(RecHist + HS_LASTBYTE_OFFSET, &read_byte, HS_LASTBYTE_LEN);
    memcpy(RecHist + HS_LASTSECOND_OFFSET, &elapsed_time, HS_LASTSECOND_LEN);
    
    SendNowPlaying(dirrecno, trackno, read_byte, elapsed_time, NOWPLAYING_FLAG_START);
    send_nowplaying = TRUE;
    continue_flag = 1;
    force_nowplaying = FALSE;
    ctrlactive_time = TimerSecond;

    while (continue_flag) {
        res = f_read(file, Mp3Buff1, DECODE_BUF_SIZE, &bytes);
        if (res != FR_OK || bytes == 0){
            break;
        }
        read_byte = file->fptr;
        sentbytes = 0;
        while (bytes > 0){
            sendbytes = (bytes > 32 ? 32 : bytes);
            while(!input(MP3_DREQ));
            output_low(MP3_DCS);
            for (i=0; i<sendbytes; i++){
                spi_write2(Mp3Buff1[sentbytes + i]);
            }
            output_high(MP3_DCS);
            bytes -= sendbytes;
            sentbytes += sendbytes;
        }

        if (/*input(MP3_DREQ) == 0 &&*/ NewBytes >= WINCTRL_CMDLENGTH){
            memcpy(WorkBuf, WifiRcvBuf + RcvIdxR, WINCTRL_CMDLENGTH);
            RcvIdxR += WINCTRL_CMDLENGTH;
            if (RcvIdxR == WIFISRBUFSIZE) RcvIdxR = 0;
            NewBytes -= WINCTRL_CMDLENGTH;
            
#if 0   // 開発時に有効にする
            fprintf(DEBUG, "Received ");
            for (i=0; i<WINCTRL_CMDLENGTH; i++){
                fprintf(DEBUG, "(%x)", WorkBuf[i]);
            }
            fprintf(DEBUG, " \r\n");
#endif
            
            csval = CalcCheckSum(WorkBuf, WINCTRL_CMDLENGTH - 1);
            if (csval != WorkBuf[WINCTRL_CMDLENGTH - 1]){
                fprintf(DEBUG, "CheckSum error \r\n");
            }
            
            if (WorkBuf[0] == WINCTRL_CMD_NEXT){                // 次の曲へ
                ret = PLAY_NEXT_MP3;
                force_nowplaying = TRUE;
                continue_flag = 0;
            } else if (WorkBuf[0] == WINCTRL_CMD_PLAYRANDOM && random == TRUE){     // ランダム再生／次のディレクトリへ進む
                ret = PLAY_OTHER_DIR;
                force_nowplaying = TRUE;
                continue_flag = 0;
            } else if (WorkBuf[0] == WINCTRL_CMD_STOPPLAY){     // 再生停止
                ret = STOP_PLAY;
                //force_nowplaying = TRUE;      // 再生停止するのでNOW_PLAYINGを送らない
                continue_flag = 0;
            } else if (WorkBuf[0] == WINCTRL_CMD_SHUTDOWN){     // ホスト切断
                SendAcceptShutdown();
                ret = SHUTDOWN;
                force_nowplaying = TRUE;
                continue_flag = 0;
            } else if (WorkBuf[0] == WINCTRL_CMD_VOLUP){        // 音量UP
                uint16_t     val = WorkBuf[1];
                if (Volume <=(val + MIN_VOLUME)){
                    Volume = MIN_VOLUME;
                } else {
                    Volume -= val;
                }
                VS1063_SetVolume(Volume);
                force_nowplaying = TRUE;
            } else if (WorkBuf[0] == WINCTRL_CMD_VOLDOWN){      // 音量DOWN
                uint16_t     val = WorkBuf[1];
                Volume += val;
                if (Volume > MAX_VOLUME){
                    Volume = MAX_VOLUME;
                }
                VS1063_SetVolume(Volume);
                force_nowplaying = TRUE;
            } else if (WorkBuf[0] == WINCTRL_CMD_FASTFORWARD){  // 曲内先送り
                memcpy(&skiptime, WorkBuf + 1, 2);
                elapsed_time = TimerSecond - timer_second_base + elapsed_time_start;
                if (bytespersec > 0 && (elapsed_time + skiptime) < total_second){
                    ofs_bk = f_tell(file);
                    ofs = ofs_bk + skiptime*bytespersec;
                    if (ofs < filesize){
                        if (f_lseek(file, ofs) == FR_OK){
                            elapsed_time_start = elapsed_time + skiptime;
                            timer_second_base = TimerSecond;
                            force_nowplaying = TRUE;
                        } else {
                            f_lseek(file, ofs_bk);
                        }
                    }
                }
            } else if (WorkBuf[0] == WINCTRL_CMD_REWIND){       // 曲内巻き戻し
                memcpy(&skiptime, WorkBuf + 1, 2);
                elapsed_time = TimerSecond - timer_second_base + elapsed_time_start;
                if (bytespersec > 0 && elapsed_time > skiptime){
                    ofs_bk = f_tell(file);
                    ofs = skiptime*bytespersec;
                    if ((ofs_bk - tagsize) > ofs){
                        ofs =  ofs_bk - ofs;
                    } else {
                        ofs = tagsize;
                    }
                    if (f_lseek(file, ofs) == FR_OK){
                        elapsed_time_start = elapsed_time - skiptime;
                        timer_second_base = TimerSecond;
                        force_nowplaying = TRUE;
                    } else {
                        f_lseek(file, ofs_bk);
                    }
                }
            } else if (WorkBuf[0] == WINCTRL_CMD_CTRLACTIVE){   // WINCTRL_CMD_NOWPLAYINGに対する返信
                ctrlactive_time = TimerSecond;
                send_nowplaying = FALSE;
            } else if (WorkBuf[0] == WINCTRL_CMD_CONFIRM){      // HOST動作確認
                UartSendData(WorkBuf, WINCTRL_CMDLENGTH);
                DebugDispSendData(WorkBuf, WINCTRL_CMDLENGTH);
            } else if (WorkBuf[0] == WINCTRL_CMD_STARTUP){      // コントローラ起動通知＆ホスト接続
                fprintf(DEBUG, "Connect\r\n");
                SendDataTimestamp();
            //} else if (WorkBuf[0] == ESPWROOM02_CMD_STOPTOPIC){   // このコマンドは使われていない
            //    SendRestartToPic();
            //    ctrlactive_time = TimerSecond;
            } else if (WorkBuf[0] == ESPWROOM02_CMD_CHECKSUMERR){   // PIC->ESP-WROOM-02のUSART送信でチェックサムエラーが発生した
                fprintf(DEBUG, "CheckSum report from ESP02\r\n");
                TXEN2 = 0;
                // 以下、WINCTRL_CMD_CTRLACTIVE待ちのタイムアウトが起きないようにする
                ctrlactive_time = TimerSecond;
                send_nowplaying = FALSE;
                // 送信
                TXEN2 = 1;
            } else {
                SendMessageUnknownCommand(WorkBuf[0]);
//                CREN2 = 0;
//                SendStopToPic();
//                WifiSrIdx = 0;      // ここまでに受信したデータをすべて無効にする
//                CREN2 = 1;
            }
            
        }
//        if (input(MP3_DREQ) == 0){
            if (continue_flag == 1 && send_nowplaying == FALSE && ((TimerSecond - ctrlactive_time) > NOWPLAYING_INTERVAL || force_nowplaying == TRUE)){
                elapsed_time = TimerSecond - timer_second_base + elapsed_time_start;
                SendNowPlaying(dirrecno, trackno, read_byte, elapsed_time, NOWPLAYING_FLAG_HALFWAY);
                send_nowplaying = TRUE;
                force_nowplaying = FALSE;
            } else if ((TimerSecond - ctrlactive_time) > NO_CTRLACTIVE_TIMEOUT){
                fprintf(DEBUG, "CTRLACTIVE stops ? \r\n");
                SendAcceptShutdown();
                ret = SHUTDOWN;
                break;
            } else if ((TimerSecond - ContinuousPlayingTime) > CONTINOUS_PLAY_LIMIT){
                fprintf(DEBUG, "Continuous playing time limit over \r\n");
                SendAcceptShutdown();
                ret = SHUTDOWN;
                break;
            }

            if (OERR2 == 1){
                SendMessageReceiveOverrun();
                CREN2 = 0;
                SendStopToPic();
                //WifiSrIdx = 0;      // ここまでに受信したデータをすべて無効にする
                CREN2 = 1;
                fprintf(DEBUG, "Overrun Error 2 \r\n");
                //UartReadAllClear();
            }
//        }
    }
    VS1063_SendZeros(2048);
    elapsed_time = TimerSecond - timer_second_base + elapsed_time_start;
//    SendNowPlaying(dirrecno, trackno, read_byte, elapsed_time, NOWPLAYING_FLAG_END);
    memcpy(RecHist + HS_LASTBYTE_OFFSET, &read_byte, HS_LASTBYTE_LEN);
    memcpy(RecHist + HS_LASTSECOND_OFFSET, &elapsed_time, HS_LASTSECOND_LEN);
    
    f_close(file);
    
    fprintf(DEBUG, "PlayOneMp3 end \r\n");
    return ret;
    
}


/*
 *  ESP-WROOM-02用　ネットワーク設定取得
 */
BOOLEAN ReadWifiConfigFile(FIL *file){
    FRESULT     res;
    UINT        br;
    BOOLEAN     ret;
    
    ret = FALSE;
    res = f_open(file, CONFIGFILE, FA_READ | FA_OPEN_EXISTING);
    fprintf(DEBUG, "f_open (%x) \r\n", res);
    if (res == FR_OK){
        res = f_read(file, RecConfig, RECLEN_CONFIGFILE, &br);
        fprintf(DEBUG, "f_read (%x) \r\n", res);
        fprintf(DEBUG, "f_read (%ld) \r\n", br);
        if (res == FR_OK && br == RECLEN_CONFIGFILE){
            ret = TRUE;
            fprintf(DEBUG, "ReadWifiConfigFile OK \r\n");
        }
        f_close(file);
    }
    return ret;
}

// ESP-WROOM-02用　ネットワーク設定データを送る
// (WINCTRL_CMDLENGTH - 2)バイトずつ送る。引数はその序数(0,1,2,...)
void SendWifiConfig(uint8_t *buf, int no){
    int     i;
    
    buf[0] = ESPWROOM02_CMD_REQCONFIG;
    buf[1] = no;
    memcpy(buf + 2, RecConfig + no*(WINCTRL_CMDLENGTH - 2), (WINCTRL_CMDLENGTH - 2));
    UartSendData(buf, WINCTRL_CMDLENGTH);
    fprintf(DEBUG, "Config File Send (%x):", no);
    for (i=0; i<WINCTRL_CMDLENGTH; i++){
        fprintf(DEBUG, "(%x)", buf[i]);
    }
    fprintf(DEBUG, " \r\n");
}

#if 0
// ESP-WROOM-02用　ネットワーク設定データを送る
// (WINCTRL_CMDLENGTH - 2)バイトずつ送る。引数はその序数(0,1,2,...)
void SendWifiConfig(FIL *file, uint8_t *buf, int no){
    FRESULT     res;
    UINT        br;
    
    res = f_open(file, CONFIGFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_lseek(file, no*(WINCTRL_CMDLENGTH - 2));
        if (res != FR_OK){
            fprintf(DEBUG, "Config File f_lseek NG (%d)\r\n", res);
        }
        res = f_read(file, buf + 2, WINCTRL_CMDLENGTH - 2, &br);
        if (res != FR_OK){
            fprintf(DEBUG, "Config File f_read NG (%d)\r\n", res);
        }
        f_close(file);
        buf[0] = ESPWROOM02_CMD_REQCONFIG;
        buf[1] = no;
        UartSendData(buf, WINCTRL_CMDLENGTH);
        fprintf(DEBUG, "Config File OK (%x)\r\n", no);
    } else {
        fprintf(DEBUG, "Config File f_open=%d\r\n", res);
    }
    
}
#endif

// 本PGの起動をESP-WROOM-02に通知
void SendPicStart(){
    
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = ESPWROOM02_CMD_PICSTART;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    fprintf(DEBUG, "send PIC start (%x, sql=%x) \r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}


/*
 *  Serial送信ブロック＆受信データクリア要求
 */
void SendStopToPic(){
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = ESPWROOM02_CMD_STOPTOPIC;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    SyncWithEspCounter = 1;
    fprintf(DEBUG, "send Serial Clear (%x, sql=%x)\r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}

/*
 *  Serial送信再開要求
 */
void SendRestartToPic(){    
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = ESPWROOM02_CMD_RESTARTTOPIC;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    fprintf(DEBUG, "send restart (%x, seq=%x)\r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}


/*
 *  ESP-WROOM-02に対してRESET要求
 */
void SendResetEsp02(){    
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = ESPWROOM02_CMD_RESETESP02;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    fprintf(DEBUG, "send reset (%x seq=%x)\r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}


/*
 *  操作待ち（再生停止中）
 */
void SendWaiting(){
    //fprintf(DEBUG, "Waiting start \r\n");
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = WINCTRL_CMD_WAITING;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    DebugDispSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    //fprintf(DEBUG, "Waiting (%x seq=%x)\r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}

/*
 *  現在再生中のディレクトリ番号と曲番号
 *      elapsed_time : 曲の先頭からの経過時間
 *      option_flag : 1=再生開始、2=再生途中、3=再生終了
 */
void SendNowPlaying(uint16_t dirrecno, uint8_t trackno, uint32_t byteposition, uint16_t elapsed_time, uint8_t option_flag){
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = WINCTRL_CMD_NOWPLAYING;
    memcpy(WifiSndBuf + 1, &dirrecno, 2);
    WifiSndBuf[3] = trackno;
    memcpy(WifiSndBuf + 4, &byteposition, 4);
    memcpy(WifiSndBuf + 8, &elapsed_time, 2);
    memcpy(WifiSndBuf + 10, &Volume, 2);
    WifiSndBuf[12] = option_flag;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
//    DebugDispSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    //fprintf(DEBUG, "NowPlaying (%x seq=%x)\r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}

/*
 *  SHUTDOWNを受け付けた
 */
void SendAcceptShutdown(){
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = WINCTRL_CMD_SHUTDOWN;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    DebugDispSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    //fprintf(DEBUG, "shutdown (%x, seq=%x)\r\n", WifiSndBuf[0], WifiSndBuf[WINCTRL_CMDLENGTH - 2]);
}

/*
 *  コマンド不正をコントローラに通知
 */
void SendMessageUnknownCommand(uint8_t cmd){
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = WINCTRL_CMD_MESSAGE;
    WifiSndBuf[1] = MESSAGE_UNKNOWN_COMMAND;
    WifiSndBuf[1] = cmd;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    DebugDispSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    //fprintf(DEBUG, "UnknownCommand (%x %x)\r\n", WifiSndBuf[0], cmd);
}

/*
 *  Receive Overrun Error をコントローラに通知
 */
void SendMessageReceiveOverrun(){
    memset(WifiSndBuf, 0, WINCTRL_CMDLENGTH);
    WifiSndBuf[0] = WINCTRL_CMD_MESSAGE;
    WifiSndBuf[1] = MESSAGE_RECEIVE_OVERRUN;
    WifiSndBuf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
    WifiSndBuf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(WifiSndBuf, WINCTRL_CMDLENGTH - 1);
    UartSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
    DebugDispSendData(WifiSndBuf, WINCTRL_CMDLENGTH);
}

#INT_EXT
void ext_isr() {
    //fprintf(DEBUG, "INT_EXT \r\n");
}

// 100msごとに発生
#INT_TIMER0
void isr_timer0(){
    uint16_t n;

    n = Timer1;             // spi_sd.c
    if (n) Timer1 = --n;
    n = Timer2;             // spi_sd.c
    if (n) Timer2 = --n;

    if (TimerStatus == 0){
        TimerCounter++;
        if (TimerCounter == 10){
            TimerCounter = 0;
            TimerSecond++;
        }
    } else {
            TimerSecond++;
    }
    
    set_timer0(Timer0InitValue);
}

/*
 *  ランダム再生順ファイル取得
 *      *rndrecno : 取得対象のランダム再生順レコード番号。実際に取得したレコード番号を取得。
 */
FRESULT ReadRandomFile(FIL *Fil, uint8_t *buf, uint16_t *rndrecno){
    FRESULT     res;
    UINT        br;
    
    res = f_open(Fil, RANDOMORDERFILE, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        fprintf(DEBUG, "rndfile f_open=%d\r\n", res);
        return res;
    }
    res = f_lseek(Fil, (uint32_t)(*rndrecno) * RECLEN_RANDOMORDER);
    if (res != FR_OK){
        *rndrecno = 0;
        res = f_lseek(Fil, (uint32_t)(*rndrecno) * RECLEN_RANDOMORDER);
    }
    res = f_read(Fil, buf, RECLEN_RANDOMORDER, &br);
    if (res != FR_OK || br != RECLEN_RANDOMORDER){
        *rndrecno = 0;
        res = f_lseek(Fil, (uint32_t)(*rndrecno) * RECLEN_RANDOMORDER);
        res = f_read(Fil, buf, RECLEN_RANDOMORDER, &br);
    }
    f_close(Fil);
    
    return FR_OK;
}

// HISTFILE読み込み
BOOLEAN ReadHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    memset(RecHist, 0, RECLEN_HISTFILE);
    res = f_open(Fil, HISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        f_read(Fil, RecHist, RECLEN_HISTFILE, &br);
        f_close(Fil);
        
        return TRUE;
    }
    return FALSE;
}

// HISTFILE保存
void SaveHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    res = f_open(Fil, HISTFILE, FA_WRITE | FA_CREATE_ALWAYS);
    if (res == FR_OK){
        res = f_write(Fil, RecHist, RECLEN_HISTFILE, &br);
        f_close(Fil);
    }
}

// ボリュームファイルからボリューム値取得
void ReadVolumeFile(FIL *file, const char *volfile){
    FRESULT     res;
    uint8_t     buf[RECLEN_VOLUME];
    UINT        br;
    
    //res = f_open(file, volfile, FA_READ | FA_OPEN_EXISTING);
    res = f_open(file, volfile, FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
        res = f_read(file, buf, RECLEN_VOLUME, &br);
        if (res == FR_OK && br == RECLEN_VOLUME){
            Volume = buf[0] | (buf[1] << 8);
            if (!(Volume >= MIN_VOLUME && Volume <= MAX_VOLUME)){
                Volume = DEFAULT_VOLUME;
            }
        } else {
            Volume = DEFAULT_VOLUME;
            if (res == FR_OK){
                res = f_lseek(file, 0);
                res = f_write(file, &Volume, RECLEN_VOLUME, &br);
            }
        }
        f_close(file);
    } else {
        Volume = DEFAULT_VOLUME;
    }
}

// ボリュームファイル保存
void SaveVolumeFile(FIL *file, const char *volfile){
    FRESULT     res;
    UINT        br;
    
    res = f_open(file, volfile, FA_WRITE | FA_CREATE_ALWAYS);
    if (res == FR_OK){
        res = f_write(file, &Volume, RECLEN_VOLUME, &br);
        f_close(file);
    }
}

// POWER-LEDの点滅
void BlinkLed(int n, uint16_t on_ms, uint16_t off_ms){
    int     i;
    
    output_low(PIN_LED);
    wait_ms(off_ms);
    for (i=0; i<n; i++){
        output_high(PIN_LED);
        wait_ms(on_ms);
        output_low(PIN_LED);
        wait_ms(off_ms);
    }
}
