/*
 * File:   vs1063.c
 * Mp3_15
 *
 * Created on 2021/03/14
 */

// VS1063 register constants
#define SCI_MODE		0x00
#define SCI_CLOCKF              0x03
#define SCI_AUDATA              0x05
#define SCI_WRAM                0x06
#define SCI_WRAMADDR            0x07
#define SCI_HDAT0               0x08
#define SCI_HDAT1               0x09
#define SCI_AIADDR              0x0A
#define SCI_VOL                 0x0B
#define SCI_AICTRL0             0x0C
#define SCI_AICTRL1             0x0D
#define SCI_AICTRL2             0x0E
#define SCI_AICTRL3             0x0F

// VS1063 SCI_MODE bits
#define SM_DIFF_B             	0x0001		// 0
#define SM_LAYER12_B          	0x0002		/* VS1063, VS1053, VS1033, VS1011 */
#define SM_RESET_B            	0x0004
#define SM_CANCEL_B           	0x0008		//3 /* VS1063, VS1053 */
#define SM_TESTS_B            	0x0020		// 5
#define SM_DACT_B             	0x0100
#define SM_SDIORD_B           	0x0200		// 9
#define SM_SDISHARE_B         	0x0400		//10
#define SM_SDINEW_B           	0x0800		// 11
#define SM_ENCODE_B           	0x1000		// 12 /* VS1063 */
#define SM_LINE1_B            	0x4000		// 14 /* VS1063, VS1053 */
#define SM_CLK_RANGE_B        	0x8000		// 15 /* VS1063, VS1053, VS1033 */

/* Parametric */
#define PAR_END_FILL_BYTE       0x1e06          /* VS1063, VS1053 */

/* WRAMADDR : bitrate mode / bitrate multiplier */
#define RQ_MODE_QUALITY 0x0000
#define RQ_MODE_VBR     0x4000
#define RQ_MODE_ABR     0x8000
#define RQ_MODE_CBR     0xc000
#define RQ_MULT_10      0x0000
#define RQ_MULT_100     0x1000
#define RQ_MULT_1000    0x2000
#define RQ_MULT_10000   0x3000
#define RQ_MP3_NO_BIT_RESERVOIR     0x0400      // (1<<10)

#define ENCODING_GAIN   (1024*9)        // original=8
#define RETRY_VS1063_MAX    10



// Encode Seting vs Stereo Mono Mode
//#define GetStereoModeOfEncSet(E)       ((EncFormatTab[(E)] & 0x07) <= 1 ? 1 : 0)        // when ADC mode (VS1063)<=1 stereo(1) else mono(0)


/*-----------------------------------------------------------------------*/
/* Hardware Reset VS1063                                                 */
/*-----------------------------------------------------------------------*/
void VS1063_HardwareReset(){
    uint16_t	mode;

//    wait_ms(250);                 // OK
    output_low(MP3_RESET);

    //SPI2 controls VS1063
    setup_spi2(SPI_MASTER | SPI_CLK_DIV_16 | SPI_SCK_IDLE_LOW | SPI_XMIT_L_TO_H | SPI_SAMPLE_AT_MIDDLE);    // for 32MHz

    wait_ms(1000);
    output_high(MP3_RESET);
	
    while(!input(MP3_DREQ));
    //mode = SM_SDINEW_B;
    VS1063_WriteSci(SCI_MODE, SM_SDINEW_B);

    while(!input(MP3_DREQ));
    mode = 0x4000;      // x2.5
    //mode = 0x2000;      // 製品版MP3のデコードには不足 (x2.0)
    //mode = 0x1000;      // too slow
    VS1063_WriteSci(SCI_CLOCKF, mode);          // CLOCKF
    
//    while(!input(MP3_DREQ));
//    VS1063_WriteSci(SCI_AIADDR, 0x50);         // Activation
    VS1063_SetVolume(Volume);
//    write_eeprom(EEPROM_LAST_VOLUME, Volume & 0x00FF);
}

/*-----------------------------------------------------------------------*/
/* Software Reset VS1063                                                 */
/*-----------------------------------------------------------------------*/
void VS1063_SoftwareReset(){
    uint16_t	mode,val;

    // Dummy read
    while(!input(MP3_DREQ));
    val = VS1063_ReadSci(SCI_MODE);

    // software reset
    while(!input(MP3_DREQ));
    mode = SM_SDINEW_B | SM_RESET_B;
    VS1063_WriteSci(SCI_MODE, mode);

    while(!input(MP3_DREQ));
    VS1063_WriteSci(SCI_HDAT0, 0xABAD);
    while(!input(MP3_DREQ));
    VS1063_WriteSci(SCI_HDAT1, 0x1DEA);
    while(!input(MP3_DREQ));
    if (VS1063_ReadSci(SCI_HDAT0) != 0xABAD || VS1063_ReadSci(SCI_HDAT1) != 0x1DEA){
    }

    VS1063_SetVolume(Volume);
//    write_eeprom(EEPROM_LAST_VOLUME, Volume & 0x00FF);
}


uint16_t VS1063_ReadSci(uint8_t addr){
    uint16_t    val = 0;

    output_low(MP3_CS);
    spi_write2(0x03);
    spi_write2(addr);
    val = spi_read2(0x00);
    val <<= 8;
    val |= spi_read2(0x00);
    output_high(MP3_CS);
    return val;
}

void VS1063_WriteSci(uint8_t addr, uint16_t val){
    output_low(MP3_CS);
    spi_write2(2);
    spi_write2(addr);
    spi_write2(val >> 8);
    spi_write2(val & 0x00FF);
    output_high(MP3_CS);
}

uint16_t VS1063_ReadWram(uint16_t addr){
    VS1063_WriteSci(SCI_WRAMADDR, addr);
    return VS1063_ReadSci(SCI_WRAM);
}

void VS1063_SetVolume(uint16_t volume){
    uint16_t    vol;

    while(!input(MP3_DREQ));
    vol = (volume << 8) | volume;
    VS1063_WriteSci(SCI_VOL, vol);          // VOL
}

void VS1063_SendZeros(uint16_t cnt){
    uint16_t    i;

    for (i=0; i<cnt; i++){
        while(!(input(MP3_DREQ)));

        output_low(MP3_DCS);
        spi_write2(0);
        output_high(MP3_DCS);
        //VS1033SENDBYTE(0);
    }
}

void LoadPluginFromFile(char *PluginFile, FIL *file) {
    int16       i = 0;
    FRESULT     res;
    uint16_t    num;
    UINT        br;

    // get number of data
    res = f_open(file, PluginFile, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        fprintf(DEBUG, "error open patch \r\n");
        return;
    }
    fprintf(DEBUG, "open patch \r\n");
    res = f_read(file, &num, 2, &br);
    if (res != FR_OK || br != 2){
        fprintf(DEBUG, "error read patch \r\n");
    }

    while (i<num) {
        //  while (i<sizeof(plugin)/sizeof(plugin[0])) {
        uint16_t addr, n, val;
        //    unsigned short addr, n, val;
        //addr = plugin[i++];
        //n = plugin[i++];
        res = f_read(file, &addr, 2, &br);
        if (res != FR_OK){
            fprintf(DEBUG, "error read patch \r\n");
        }
        i++;
        res = f_read(file, &n, 2, &br);
        if (res != FR_OK){
            fprintf(DEBUG, "error read patch \r\n");
        }
        i++;
        if (n & 0x8000U) { /* RLE run, replicate n samples */
            n &= 0x7FFF;
            //val = plugin[i++];
            res = f_read(file, &val, 2, &br);
            if (res != FR_OK){
                fprintf(DEBUG, "error read patch \r\n");
            }
            i++;
            while (n--) {
                while(!input(MP3_DREQ));
                VS1063_WriteSci(addr, val);
                //fprintf(DEBUG, "write patch \r\n");
            }
        } else {           /* Copy run, copy n samples */
            while (n--) {
                //val = plugin[i++];
                res = f_read(file, &val, 2, &br);
                if (res != FR_OK){
                    fprintf(DEBUG, "error read patch \r\n");
                }
                i++;
                while(!input(MP3_DREQ));
                VS1063_WriteSci(addr, val);
                //fprintf(DEBUG, "write patch \r\n");
            }
        }
    }

    f_close(file);
    fprintf(DEBUG, "end read patch \r\n");

}
